#!/usr/bin/python3
from configparser import NoOptionError
from configparser import ConfigParser as SafeConfigParser
import os.path
import os
import shutil
import logging
import threading
import subprocess
from SystemUpdater.Core.utils import UpdaterKconf2Err

class UpgradeConfig(SafeConfigParser):
    # 备份还原之前保存的配置
    UPDATE_NOT_RESOTRE_DIR = "/var/lib/kylin-system-updater/remain"

    def __init__(self, datadir="/var/lib/kylin-system-updater/", name="system-updater.conf",defaults_conf=None):
        SafeConfigParser.__init__(self)
        self.datadir = datadir
        self.maincfg = os.path.join(datadir, name)

        self.config_files = []
        # 增加的默认配置，先读取默认配置加载配置项，最终写入到主配置内
        if defaults_conf:
            self.config_files.append(os.path.join(datadir, defaults_conf))
        self.config_files.append(self.maincfg)
        
        # 检查是否存在备份的配置，存在时直接移动到默认位置
        self.remail_conf = os.path.join(self.UPDATE_NOT_RESOTRE_DIR, name)
        if os.path.exists(self.remail_conf):
            shutil.move(self.remail_conf,self.maincfg)

        # 检查配置文件是否存在
        for conf in self.config_files:
            if not os.path.exists(conf):
                logging.warning("Check: Conf(%s) not exits...",conf)
                # self.config_files.remove(conf)

        self.read(self.config_files, encoding='iso-8859-1')
        logging.info("Initialize Upgrade ConfigFile(%s) to success",str(self.config_files))

    def reRead(self):
        self.read(self.config_files, encoding='iso-8859-1')

    def setValue(self, section, option, value=None,is_write = True):
        logging.info("SetValue Section:%s Option:%s Value:%s",section, option, value)
        try:
            self.reRead()
            self.set(section, option,value)
        except Exception as e:
                logging.error("Error: setValue section:%s option:%s value:%s",section, option, value)
                logging.error(str(e))
                return False
        if is_write == True:
            with open(self.maincfg, 'w+') as configfile:
                self.write(configfile)
        return True

    def getWithDefault(self, section, option, default,re_read=False):
        try:
            if re_read == True:
                self.reRead()

            if type(default) == bool:
                return self.getboolean(section, option)
            elif type(default) == float:
                return self.getfloat(section, option)
            elif type(default) == int:
                return self.getint(section, option)
            return self.get(section, option)
        except Exception as e:
            logging.warning(str(e))
            return default

    def getlist(self, section, option):
        try:
            tmp = self.get(section, option)
        except Exception as e:
            logging.error(str(e))
            return []
        items = [x.strip() for x in tmp.split(" ")]
        if '' in items and len(items):
            return []
        return items

    def setListValue(self, section, option, value=None,is_write = True):
        tmp = str(value).replace('[', '').replace(']', '')
        tmp = tmp.replace("'", '').replace(',', '')
        try:
            self.set(section, option,tmp)
        except Exception as e:
            logging.error("setListValue section:%s option:%s",section, option)
            logging.error(str(e))
            return
        if is_write == True:
            with open(self.maincfg, 'w+') as configfile:
                self.write(configfile)

    def getListFromFile(self, section, option):
        try:
            filename = self.get(section, option)
        except NoOptionError:
            return []
        p = os.path.join(self.datadir, filename)
        if not os.path.exists(p):
            logging.error("getListFromFile: no '%s' found" % p)
        with open(p) as f:
            items = [x.strip() for x in f]
        return [s for s in items if not s.startswith("#") and not s == ""]

    # 还原之前保留配置
    def remain_conf(self):
        shutil.copy(self.maincfg, self.remail_conf)

    def optionxform(self, optionstr: str) -> str:
        return optionstr

class UpgradeKConf2Manager():
    APPLICATION_NAME = 'kylin-update-frontend'
    GROUP_TO_LISTEN  = [
        'kylin-update-frontend.notification',
        'kylin-update-frontend.downloadlimit',
        'kylin-update-frontend.autodownload'
        ]

    def __init__(self, dbusCtrl = None, skip_listen = False):
        self.dbusControlor  = dbusCtrl
        self.pty_lock       = threading.Lock()
        self.master_fds     = []
        self.should_exit    = False

        if not skip_listen:
            self.conf2_listen_keys_changed()
    
    def quitKConf2Manager(self):
        self.should_exit = True

    def _parse_kconf2_output(self, output):
        updater_kconf2_childrens = []

        if "Wrong schema id" != str(output).strip():
            updater_kconf2_childrens = output.splitlines()

        return updater_kconf2_childrens

    def _parse_kconf2_value(self, output):
        return str(output).strip()
    
    def _get_updater_kconf2_childrens(self):
        updater_kconf2_childrens = None

        args = ["kconf2", "list-children", self.APPLICATION_NAME]

        process = subprocess.Popen(" ".join(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output, error = process.communicate()
        if not error:
            updater_kconf2_childrens = self._parse_kconf2_output(output.decode('utf-8'))
        else:
            logging.error("UpgradeConfigKConf2: %s",str(error))

        return updater_kconf2_childrens

    def conf2_listen_keys_changed(self):
        try:
            if not os.path.exists("/usr/bin/kconf2"):
                logging.error("conf2_listen_keys_changed: kconf2 not found.")
                return

            updater_kconf2_childrens = self._get_updater_kconf2_childrens()
            for ikc in updater_kconf2_childrens:
                if ikc in self.GROUP_TO_LISTEN:
                    monitor_thread = threading.Thread(target=self._listen_key_changed, args=(ikc,))
                    monitor_thread.start()

        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))

    def _listen_key_changed(self, schema):
        try:
            import pty
            import select

            master_fd = -1
            slave_fd  = -1

            master_fd, slave_fd = pty.openpty()

            args = ["kconf2", "monitor", schema]
            process = subprocess.Popen(args, stdout=slave_fd, stderr=slave_fd, text=True)
            # logging.warning("kconf2 monitor [%s %d]", schema, threading.get_ident())

            os.close(slave_fd)

            self.master_fds.append(master_fd)

            while not self.should_exit:
                rlist, _, _ = select.select([master_fd], [], [], 1)

                if master_fd in rlist:
                    output = os.read(master_fd, 1024).decode('utf-8')
                    if output:
                        with self.pty_lock:
                            # logging.warning("[%s %d]:%s", schema, threading.get_ident(), output.strip())

                            change_details = [part.strip() for part in output.strip().split('changed to')]

                            if schema == 'kylin-update-frontend.notification':

                                if "enable" == str(change_details[0]):
                                    self.dbusControlor.AutocheckStatusChanged(
                                                        True if str(change_details[-1]).lower() == 'true' else False)

                            elif schema == 'kylin-update-frontend.downloadlimit':

                                if "enable" == str(change_details[0]):
                                    downloadlimitValue = self.conf2_get_key_value(schema, 'value')
                                    self.dbusControlor.DownloadStatusChanged(
                                                        True if str(change_details[-1]).lower() == 'true' else False,
                                                        int(downloadlimitValue))
                                
                                elif "value" == str(change_details[0]):
                                    downloadspeedLimitStatus = self.conf2_get_key_value(schema, 'enable')
                                    self.dbusControlor.DownloadStatusChanged(
                                                        True if downloadspeedLimitStatus.lower() == 'true' else False,
                                                        int(str(change_details[-1])) if str(change_details[-1]).isdigit() else int(0))

                            elif schema == 'kylin-update-frontend.autodownload':

                                if "enable" == str(change_details[0]):
                                    self.dbusControlor.AutoUpgradeStatusChanged(
                                                        True if str(change_details[-1]).lower() == 'true' else False)

            process.terminate()
        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))

    def conf2_get_key_value(self, schema, key):
        updater_kconf2_value = ''
        
        if not os.path.exists("/usr/bin/kconf2"):
            raise UpdaterKconf2Err()

        try:
            args = ["kconf2", "get", schema, key]

            process = subprocess.Popen(" ".join(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, error = process.communicate()
            if not error:
                updater_kconf2_value = self._parse_kconf2_value(output.decode('utf-8'))
            else:
                logging.warning(str(error))
            logging.info("UpgradeConfigKConf2: %s:%s -> %s.", schema, key, updater_kconf2_value)
        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))

        return updater_kconf2_value

    def conf2_set_key_value(self, schema, key, value):
        setStatus = True
        try:
            args = ["kconf2", "set", schema, key, value]

            process = subprocess.Popen(" ".join(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, error = process.communicate()
            logging.info("UpgradeConfigKConf2: %s:%s -> %s.", schema, key, value)
        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))
            setStatus = False

        return setStatus


if __name__ == "__main__":
    pass