# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
import os
import sys
import time
import dbus
import logging
import dbus.service
import traceback
from .Core.errors import *
from .Core.enums import *
from gettext import gettext as _
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)
from .Core.loop import mainloop
from .UpgradeStrategiesDbus import UpgradeStrategiesDbusController,UPDATER_DBUS_INTERFACE,UPDATER_DBUS_PATH,UPDATER_DBUS_SERVICE
from .backendnext import InstallBackend,get_backend
from SystemUpdater.Core.space_rules import check_free_space
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig,UpgradeKConf2Manager
from .Core.Database import Sqlite3Server

STRATEGY_IDLE_INTERVAL = 2*60
STRATEGY_IDLE_TIMEOUT = 6*60
STRATEGIES_UPDATING_CACHE = "/tmp/upgrade-strategies-update-cache"

class UpgradeStrategies():
    def __init__(self,options):
        try:
            self.options = options
            self.cache = None
            self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
            self.sqlite3_server = Sqlite3Server(self)
            
            #dbus
            self.dbusController = self._setup_dbus()

            self.bus = dbus.SystemBus()
            #config
            self.uuconfigs = UpgradeConfig(datadir = "/var/lib/unattended-upgrades/", name = "unattended-upgrades-policy.conf")
            self.configs_uncover = UpgradeConfig(defaults_conf="system-updater-defaults.conf")
            self.configs_cover = UpgradeConfig(name = "system-updater-coverable.conf")
            #策略配置接口的超时退出机制
            # kconf2配置加载
            self.updater_kconf2 = UpgradeKConf2Manager(self.dbusController)

            self.strategy_timestamp = 0

            self.check_update_cache_status()
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
    
    def run(self):
        """Start the daemon and listen for calls."""
        logging.info("Waiting for calls...")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbusController.Quit(None)

    def _setup_dbus(self):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        bus = dbus.SystemBus()
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            logging.info("Initiate dbus success ...")
            return  UpgradeStrategiesDbusController(self, bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpgradeStrategiesDbusController(self, bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)
    
    def _check_strategy_inactivity(self):
        logging.info("Checking for inactivity in Strategies daemon ...")
        timestamp = self.strategy_timestamp
        if timestamp == 0:
            self.strategy_timestamp = time.time()
            return True
        #超时退出
        if self.strategy_timestamp != 0 and  time.time() - self.strategy_timestamp > STRATEGY_IDLE_TIMEOUT:
            logging.warning("Quitting due to inactivity")
            self.dbusController.Quit(None)
            return False
        return True

    #更新important.list的本次升级的列表
    def _update_important(self,reply_handler,error_handler):
        def _update_important_reply(retval):
            if bool(retval) == False:
                reply_handler()
            else:
                reply_handler()
        def _update_important_error(retval):
            logging.error(str(retval))
            reply_handler()
        
        try:
            obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')

            methods = dbus.Interface(obj, dbus_interface='org.freedesktop.DBus.Introspectable').Introspect()
            interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
            if "checkUpdatesInStages" in methods:
                self.dbusController.UpdateCacheStatusChanged(5,_("Updating Source Template")+"(checkUpdatesInStages)")
                interface.checkUpdatesInStages(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
            else:
                self.dbusController.UpdateCacheStatusChanged(5,_("Updating Source Template")+"(updateSourceTemplate)")
                interface.updateSourceTemplate(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
        except Exception as e:
            logging.warning(str(e))
            reply_handler()

    def start_update_cache(self):
        def _update_important_success():
            self.start_update_backend()
            return
        try:
            # 创建文件标记正在更新检测
            with open(STRATEGIES_UPDATING_CACHE, "w") as f:
                pass

            check_free_space()
        except UpdateBaseError as excep:
            error_code = excep.code
            self.dbusController.UpdateCacheFinishedNext(False, [''], get_error_num_from_enum(error_code),\
                                                        get_error_string_from_enum(error_code)+get_error_description_from_enum(error_code))
            return
        except Exception as e:
            logging.error(e)

        self._update_important(_update_important_success,_update_important_success)

    def start_update_backend(self):
        # self.check_config_aptdeamon()
        update_backend = get_backend(self, InstallBackend.ACTION_UPDATE)
        update_backend.start()

    def check_update_cache_status(self):
        try:
            # 检查上次更新检测是否未正常结束
            if os.path.exists(STRATEGIES_UPDATING_CACHE):
                self.dbusController.UpdateCacheFinishedNext(False, [''], get_error_num_from_enum(ERROR_STRATEGIES_RESTART),\
                                            get_error_string_from_enum(ERROR_STRATEGIES_RESTART)+_(get_error_description_from_enum(ERROR_STRATEGIES_RESTART)))
        except Exception as e:
            logging.error("check_update_cache_status error: %s.",str(e))
