#!/usr/bin/python3
import os
import dbus
import json
import dbus.service
import logging
import subprocess
from .backendnext import InstallBackend
from gettext import gettext as _
from .Core.loop import mainloop
from SystemUpdater.Core.utils import get_proc_from_dbus_name, refresh_system_language, UpdaterKconf2Err
from SystemUpdater.Core.DbusControl import DbusControl
from .UpdatesAvailable import UpdatesAvailable

UPDATER_DBUS_INTERFACE = 'com.kylin.UpgradeStrategies.interface'
UPDATER_DBUS_PATH = '/com/kylin/UpgradeStrategies'
UPDATER_DBUS_SERVICE = 'com.kylin.UpgradeStrategies'
RUN_UNATTENDED_UPGRADE = '/var/run/unattended-upgrades.pid'
SYSTEM_VERSION = '/etc/kylin-version/kylin-system-version.conf'
STRATEGIES_UPDATING_CACHE = "/tmp/upgrade-strategies-update-cache"

#颜色设置
COLORLOG_SUFFIX = "\033[0m"

# Define some foreground colors
BLACK = 30
RED = 31
GREEN = 32
YELLOW = 33
BLUE = 34
MAGENTA = 35
CYAN = 36
WHITE = 37

#字体颜色
FRONT_COLOR_SEQ = "\033[1;%dm"
#背景颜色
BACK_COLOR_SEQ = "\033[%d;1m"

COLORLOG_PREFIX     = FRONT_COLOR_SEQ % GREEN                                                                 
COLORMETHOR_PREFIX  = FRONT_COLOR_SEQ % CYAN

UU_UPGRADE_MODE_AUTOMATIC_DOWNLOAD = 0
UU_UPGRADE_MODE_MANUAL = 1
UU_UPGRADE_MODE_AUTOMATIC_INSTALL = 2
UU_UPGRADE_MODE_BEFORE_SHUTDOWN = 3
UU_UPGRADE_MODE_WORKTIME = 4

#dbus 建立
class UpgradeStrategiesDbusController(dbus.service.Object):
    """ this is a helper to provide the UpdateManagerIFace """

    P2P_DEDAULT_PATH = "/etc/default/apt-p2p"
    RETURN_SUCCESS_CODE = 0
    RETURN_SUCCESS_DESC = ""

    RETURN_UNKNOWN_CODE = -1
    RETURN_UNKNOWN_DESC = ""

    RETURN_BUSY_STATE = 1
    RETURN_BUSY_DESC = "In the process of updating or Upgrading..."

    def __init__(self, parent, bus_name,
                object_path=UPDATER_DBUS_PATH):
        dbus.service.Object.__init__(self, bus_name, object_path)
        self.parent = parent
        self.bus = dbus.SystemBus()
        self.dbus_control = DbusControl(self.bus,parent.options.debug)
        self.transaction = None

    def __check_change__(self, _config = None, _section = "", _option = "", _value = ""):
        if _config == None:
            return False
        if _value == _config.getWithDefault(_section,_option,_value," "):
            return True
        return False

    #设置aptdeamon的环境变量
    def set_aptdeamon_environ(self,key,value):
        try:
            logging.info("Set aptdeaom environment variables %s = %s...",key,value)
            obj = self.bus.get_object('org.debian.apt', '/org/debian/apt')
            interface = dbus.Interface(obj, dbus_interface='org.debian.apt')
            retval = interface.SetEnviron(key,value)
            return retval
        except Exception as e:
            logging.error(str(e))
            return False

    def kconf2_get_downloadlimit_value(self):
        downloadspeedLimitStatus    = False
        downloadlimitValue          = "0"

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadlimit', 'enable')
            downloadspeedLimitStatus = False if kconf2_key_value.lower() == 'false' else True
            
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadlimit', 'value')
            downloadlimitValue = str("0") if not kconf2_key_value.isdigit() else str(kconf2_key_value)
        except Exception as e:
            logging.error("kconf2_get_downloadlimit_value err: %s.", str(e))
            downloadlimitValue = "err"

        return downloadspeedLimitStatus, downloadlimitValue

    def kconf2_set_downloadlimit_value(self, _speed_value):
        setResult = int(0)

        try:
            speed_value = int(_speed_value)

            if speed_value < 0:
                return int(1)

            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.downloadlimit', 
                                                                'value',
                                                                str(speed_value)) else int(1)
        except Exception as e:
            logging.error("SetDownloadLimit: %s.", str(e))
            setResult = int(1)

        return setResult

    def kconf2_set_downloadlimit_enable(self, status):
        setResult = int(0)

        RATE_CONFIG_FILE = "/var/cache/kylin-system-updater/rate_limites.json"
        rate_config = {
            "policy_limits": False,
            "rate_limits": [],
            "default_rate": "0"
        }

        try:
            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.downloadlimit', 
                                                                'enable', 
                                                                'true' if status else 'false') else int(1)
            if setResult == int(0):
                if status:
                    kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadlimit', 'value')
                    downloadlimitValue = int(0) if not kconf2_key_value.isdigit() else int(kconf2_key_value)

                    rate_config["default_rate"] = str(downloadlimitValue)
                else:
                    if os.path.exists("/etc/apt/apt.conf.d/80apt-download"):
                        os.remove("/etc/apt/apt.conf.d/80apt-download")

            with open(RATE_CONFIG_FILE, 'w') as config_file:
                json.dump(rate_config, config_file, indent=4)

        except Exception as e:
            logging.error("SetDownloadLimitStatus: %s.", str(e))
            setResult = int(1)

        return setResult

    def kconf2_set_downloadsettings_period(self, _period):
        setResult = int(0)

        try:
            period = int(_period)

            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.downloadsettings', 
                                                                'period', 
                                                                str(period)) else int(1)

            if setResult == int(0):
                self.AutoupgradePeriodChanged(period)

        except Exception as e:
            logging.error("SetAutoupgradePeriod: %s.", str(e))
            setResult = int(1)

        return setResult

    def kconf2_set_downloadsettings_designated(self, status, designatedtime):
        setResult = int(0)

        if ":" not in designatedtime or "-" not in designatedtime:
            return int(1)

        try:
            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.downloadsettings', 
                                                                'designated',
                                                                'true' if status else 'false') else int(1)

            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.downloadsettings', 
                                                                'designatedtime',
                                                                str(designatedtime)) \
                                else int(1)

            if setResult == int(0):
                self.AutoupgradeDetailChanged(status, str(designatedtime))

        except Exception as e:
            logging.error("SetAutoupgradeDetail: %s.", str(e))
            setResult = int(1)

        return setResult

    @dbus.service.method(UPDATER_DBUS_INTERFACE,
                        in_signature="", out_signature="",
                        sender_keyword="sender")
    def Quit(self, sender):
        """Request a shutdown of the daemon."""
        if sender:
            sender_name = self.dbus_control.check_sender(sender)
        #如果在下载就请求 取消
        logging.info("UpgradeStrategies service quitting...")
        self.parent.updater_kconf2.quitKConf2Manager()
        mainloop.quit()
        logging.debug("Exit")

    ## dbus接口: 开启或关闭预下载功能
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='is',sender_keyword='sender')
    def ChangingP2PStatus(self,_status,sender = None):
        status = str(_status)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' ChangingP2PStatus, _status = %s , sender name: %s',status,sender_name)
        
        if os.path.exists(self.P2P_DEDAULT_PATH):
            if status == "enable":
                with open(self.P2P_DEDAULT_PATH, 'w+') as configfile:
                    configfile.write("#enable=true\n")
                #第一次进行检查是否已经开启
                args = ["systemctl","is-enabled","apt-p2p.service"]
                p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
                if p.returncode != 0:
                    logging.info("Apt-p2p service is not runing and will be enable...")
                    #第二次进行重启
                    args = ["systemctl","enable","apt-p2p.service"]
                    p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
                    if p.returncode == 0:
                        logging.info("Service Enable Execute successfully")
                        #第三次进行重启
                        args = ["systemctl","restart","apt-p2p.service"]
                        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
                        if p.returncode == 0:
                            logging.info("Restart Execute successfully")
                            return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC
                        else:
                            logging.error(str(p.stdout))
                            logging.error("Failed to execute reboot")
                            return self.RETURN_UNKNOWN_CODE,str(p.stdout)
                    else:
                        logging.error(str(p.stdout))
                        logging.error("Failed to execute enable")
                        return self.RETURN_UNKNOWN_CODE,str(p.stdout)
                else:
                    logging.info("Apt-p2p service has been enabled and not need to reabled...")
                    return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC

            elif status == "disable":
                with open(self.P2P_DEDAULT_PATH, 'w+') as configfile:
                    configfile.write("enable=false\n")

                args = ["systemctl","is-enabled","apt-p2p.service"]
                p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
                if p.returncode == 0:
                    logging.info("Apt-p2p service is runing and will be disable...")
                    args = ["systemctl","disable","apt-p2p.service"]
                    p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
                    if p.returncode == 0:
                        logging.info("Service disable Execute successfully")

                        args = ["systemctl","stop","apt-p2p.service"]
                        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)

                        if p.returncode == 0:
                            logging.info("Stop Execute successfully")
                            return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC
                        else:
                            logging.error(str(p.stdout))
                            logging.error("Failed to execute Stop")
                            return self.RETURN_UNKNOWN_CODE,str(p.stdout)
                    else:
                        logging.error(str(p.stdout))
                        logging.error("Failed to execute disable")
                        return self.RETURN_UNKNOWN_CODE,str(p.stdout)
                else:
                    logging.info("Apt-p2p service has been disabled and not need to redisabled...")
                    return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC
            else:
                logging.warning("error: input value _status=%s",status)
        else:
            logging.warning("apt-p2p function is not install...")

    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='b', out_signature='b',sender_keyword='sender')
    def SetUpgradeStrategyState(self,state,sender=None):
        state=bool(state)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetUpdateStrategyState, upgrade strategy state is %r, sender name: %s .',state,sender_name)
        try:
            if state:
                if not self.__check_change__(self.parent.uuconfigs, "updateStrategiesManager", "strategiesState", "True"):
                    self.parent.uuconfigs.setValue("updateStrategiesManager", "strategiesState", "True", True)                    
                    self.StrategyChanged("strategiesState","True")            
            else:
                if not self.__check_change__(self.parent.uuconfigs, "updateStrategiesManager", "strategiesState", "False"):
                    self.parent.uuconfigs.setValue("updateStrategiesManager", "strategiesState", "False", True)
                    self.StrategyChanged("strategiesState","False")
        except Exception as e:
            logging.error(str(e))
            return False
        return True
    
        ## 设置更新周期
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='b',sender_keyword='sender')
    def SetUpgradeMode(self, mode, sender = None):
        _mode = str(mode)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetUpgradeMode, mode: %s , sender:%s .',_mode,sender_name)                        
        try:
            if not self.__check_change__(self.parent.uuconfigs, "updateStrategiesManager", "installType", _mode):
                self.parent.uuconfigs.setValue("updateStrategiesManager", "installType", _mode, True)
                self.StrategyChanged("installType",_mode)
        except Exception as e:
            logging.error(str(e))
            return False
        return True
    
    ## dbus接口: 开启或关闭预下载功能
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='bs', out_signature='b',sender_keyword='sender')
    def SetPreDownloadState(self, _state, _time, sender = None):
        state = bool(_state)
        time  = str(_time)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetPreDownloadState, state is %r, time: %s, sender name: %s .',state,time,sender_name)
        try:
            if state:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "preDownload", "on"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "preDownload", "on", True)
                    self.PropertyChanged("preDownload","on")
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "preDownloadTime", time):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "preDownloadTime", time, True)
                    self.PropertyChanged("preDownloadTime",time)
            else:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "preDownload", "off"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "preDownload", "off", True)
                    self.PropertyChanged("preDownload","off")
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    ## 设置更新周期
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='i', out_signature='b',sender_keyword='sender')
    def SetUpdateDays(self, days, sender = None):
        _days = int(days)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetUpdateDays, days: %d , sender:%s .'\
                        ,_days,sender_name)
        try:
            if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "updateDays", _days):
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "updateDays", str(_days), True)
                self.PropertyChanged("updateDays",str(_days))

            self.kconf2_set_downloadsettings_period(_days)
        except Exception as e:
            logging.error(str(e))
            return False
        return True
    
    # 设置自动更新时间
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='bs',sender_keyword='sender')
    def SetAutoUpgradeRandomRange(self,randomRange,sender=None):
        _randomRange = str(randomRange)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutoUpgradeRandomRange will be set value %s, sender: %s .',\
                            _randomRange,sender_name)
        try:
            if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "randomRange", _randomRange):
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "randomRange", _randomRange, True)
                self.PropertyChanged("randomRange",_randomRange)
        except Exception as e:
            logging.error(str(e))
        return True,"success"

    ## 设置是否开启自动重启,以及设置自动重启的时间
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='bs', out_signature='b',sender_keyword='sender')
    def SetAutomaticReboot(self, status, reboot_time, sender = None):
        _state = bool(status)
        _reboot_time  = str(reboot_time)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetAutomaticReboot, status is %r, reboot_time: %s, sender name: %s .'\
                        ,_state,_reboot_time,sender_name)
        try:
            if _state:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "automaticReboot", "on"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "automaticReboot", "on", True)
                    self.PropertyChanged("automaticReboot","on")
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "automaticRebootTime", _reboot_time):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "automaticRebootTime", _reboot_time, True)
                    self.PropertyChanged("automaticRebootTime",_reboot_time)
            else:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "automaticReboot", "off"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "automaticReboot", "off", True)
                    self.PropertyChanged("automaticReboot","off")
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    ## dbus接口: 开启关闭自动更新功能
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='b', out_signature='b',sender_keyword='sender')
    def SetAutoUpgradeState(self, _state,sender = None):
        state = bool(_state)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetAutoUpgradeState, state is %r ...',state)
        try:
            if state:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "autoUpgradeState", "on"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "autoUpgradeState", "on", True)
                    self.parent.sqlite3_server.insert_into_display("autoupdate_allow", "true")
                    self.PropertyChanged("autoUpgradeState","on")
                self.ButtonStatusChange("autoUpgradeStatus", "true")
            else :
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "autoUpgradeState", "off"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "autoUpgradeState", "off", True)
                    self.parent.sqlite3_server.insert_into_display("autoupdate_allow", "false")
                    self.PropertyChanged("autoUpgradeState","off")
                self.ButtonStatusChange("autoUpgradeStatus", "false")
            
            self.parent.updater_kconf2.conf2_set_key_value('kylin-update-frontend.autodownload', 
                                                            'enable', 
                                                            'true' if state else 'false')
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    ## dbus接口: 设置自动更新策略
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='is', out_signature='b',sender_keyword='sender')
    def SetAutoUpgradeMode(self, mode, time,sender = None):
        _time = str(time)
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetAutoUpgradeMode, mode is %s, time is %s ...',mode,_time)
        try:
            if mode == UU_UPGRADE_MODE_AUTOMATIC_DOWNLOAD:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "downloadMode", "timing"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadMode", "timing", True)
                    self.PropertyChanged("downloadMode","timing")

                    self.kconf2_set_downloadsettings_designated(True, _time)
                    # designatedtime:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "downloadTime", str(_time)):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadTime", str(_time), True)
                    self.PropertyChanged("downloadTime",str(_time))
                self.ButtonStatusChange("autoUpgradeTime", str(_time))
            elif mode == UU_UPGRADE_MODE_AUTOMATIC_INSTALL:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "installMode", "timing"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "installMode", "timing", True)
                    self.PropertyChanged("installMode","timing")
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "installTime", str(_time)):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "installTime", str(_time), True)
                    self.PropertyChanged("installTime",str(_time))
            elif mode == UU_UPGRADE_MODE_BEFORE_SHUTDOWN:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "installTime", "bshutdown"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "installMode", "bshutdown", True)
                    self.PropertyChanged("installMode","bshutdown")
            elif mode == UU_UPGRADE_MODE_MANUAL:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "downloadMode", "manual"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadMode", "manual", True)
                    self.PropertyChanged("downloadMode","manual")
            elif mode == UU_UPGRADE_MODE_WORKTIME:
                if not self.__check_change__(self.parent.uuconfigs, "autoUpgradePolicy", "downloadMode", "manual"):
                    self.parent.uuconfigs.setValue("autoUpgradePolicy", "workTime", time, True)
                    self.PropertyChanged("workTime",time)
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    # # dbus接口：改变apt下载速度
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='sb', out_signature='is',sender_keyword='sender')
    def SetDownloadspeedMax(self, speed, set,sender = None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetDownloadspeedMax, speed:%s, set:%r, sender name: %s .'%(speed, set, sender_name))

        RATE_CONFIG_FILE = "/var/cache/kylin-system-updater/rate_limites.json"
        rate_config = {
            "policy_limits": False,
            "rate_limits": [],
            "default_rate": "0"
        }

        if set:
            # FIX:安全漏洞 仅下载速度字符串全为数字时才可使用
            speed = str(speed).strip()
            if not speed.isdigit():
                return self.RETURN_UNKNOWN_CODE,""
            
            rate_config["default_rate"] = str(speed)
            try:
                with open(RATE_CONFIG_FILE, 'w') as config_file:
                    json.dump(rate_config, config_file, indent=4)
                
                self.parent.sqlite3_server.insert_into_display("download_limit","true")
                self.parent.sqlite3_server.insert_into_display("download_limit_value",str(speed))
                #发送信号
                self.ButtonStatusChange("speed" , str(speed))
            except Exception as e:
                logging.error(e)
                return self.RETURN_UNKNOWN_CODE,str(e)

            self.kconf2_set_downloadlimit_value(int(speed))
            self.kconf2_set_downloadlimit_enable(True)
        else:
            with open(RATE_CONFIG_FILE, 'w') as config_file:
                json.dump(rate_config, config_file, indent=4)
            self.parent.sqlite3_server.insert_into_display("download_limit","false")
            self.ButtonStatusChange("speed", "0")

            self.kconf2_set_downloadlimit_value(int(speed))
            self.kconf2_set_downloadlimit_enable(False)
        return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC

    # # dbus接口：获取apt下载速度
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='bs',sender_keyword='sender')
    def GetDownloadspeedLimitValue(self,sender = None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetDownloadspeedLimitValue sender: %s .', sender_name)

        download_limit          = False
        download_limit_value    = "0"
        try:
            download_limit, download_limit_value = self.kconf2_get_downloadlimit_value()
            if download_limit_value == "err":
                download_limit = self.parent.sqlite3_server.select_from_display("download_limit")
                if download_limit == "true":
                    download_limit_value = self.parent.sqlite3_server.select_from_display("download_limit_value")
                    return True,str(download_limit_value)
                else:
                    return False,str("0")
        except Exception as e:
            logging.error("GetDownloadspeedLimitValue error:%s", e)
        
        return download_limit, str(download_limit_value)

    # 是否允许关机前更新
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='ss', out_signature='bs', sender_keyword='sender')
    def UnattendedUpgradeValue(self, operation, value="false", sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UnattendedUpgradeValue sender:%s ', sender_name)
        if operation.lower() != "get" and operation.lower() != "set":
            return False, 'Please input [\"set\", \"value\"] to set. \nor [\"get\"] to get whether updates are allowed before shutdown.'
        if operation == "set":
            try:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UnattendedUpgradeValue is going to %s [allow_unattended_upgrades_shutdown] value to %s.'%(operation,value))
                self.parent.sqlite3_server.insert_into_display("allow_unattended_upgrades_shutdown", value.lower())
            except Exception as e:
                logging.error(str(e))
                return False,str(e)
        else:
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UnattendedUpgradeValue is going to %s [allow_unattended_upgrades_shutdown] value.'%(operation))
            try:
                value = self.parent.sqlite3_server.select_from_display("allow_unattended_upgrades_shutdown")
                logging.info("[allow_unattended_upgrades_shutdown] value is %s."%(value))
            except Exception as e:
                logging.error(str(e))
                return False,str(e)
        return True,"success"

    # 获取数据库值
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='bss', out_signature='s', sender_keyword='sender')
    def GetSetDatabaseInfo(self, gs, table, field,sender=None):
        Text = 'NULL'
        sender_name = self.dbus_control.check_sender(sender)
        try:
            if gs: #get 
                if table == 'display':
                    Text = self.parent.sqlite3_server.select_from_display(str(field))
                    logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetDatabaseInfo Table:%s Field:%s Text:%s',table,field,Text)
            else: #set 
                if table == 'display' and "=" in field:
                    field, value = str(field).split("=")
                    logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetDatabaseInfo Table:%s Field:%s', table, field)
                    self.parent.sqlite3_server.insert_into_display(field, value)
                    return "success"
        except Exception as e:
            logging.error(str(e))
            return Text
        return Text

    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='s', sender_keyword='sender')
    def GetUpdateStatusDict(self,key,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        try:
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetUpdateStatusDict key:%s',key)
            # 更新关机安装相应的安装数据
            with open(UpdatesAvailable.UPDATE_AVAILABLE_DATA, 'r') as file:
                data = json.load(file)
            
            value = str(data[key])
            logging.info("update status value:%s",value)
            return value
        except Exception as e:
            logging.error(str(e))
            return ""

    ## dbus接口: 发送立即更新的信号
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='b', sender_keyword='sender')
    def AutoUpgradeAllNow(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' AutoUpgradeAllNow ...')
        try:
            self.UpgradeAllNow()
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    #设置数据库配置信息
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='ss',out_signature='b',sender_keyword='sender')
    def DatabaseInfoSet(self,field_name,field_value,sender=None):
        Status = False
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetDatabaseInfo,field_name:%s,field_value:%s,caller:%s .',\
                field_name,field_value,sender_name)
        Status = self.parent.sqlite3_server.insert_into_display(field_name,field_value)
        return bool(Status)

    #数据库获取配置信息
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='s',out_signature='s',sender_keyword='sender')
    def DatabaseInfoGet(self,field_name,sender=None):
        field_value = ''
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetDatabaseInfo field_name:%s caller:%s',field_name,sender_name)
        field_value = self.parent.sqlite3_server.select_from_display(str(field_name))
        logging.info("Get field_value:%s",field_value)
        return field_value

    #更新的dbus
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='is',sender_keyword='sender')
    def UpdateCache(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.now_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('UpdateCache In the process of Updating or Upgrading...')
                return self.RETURN_BUSY_STATE,self.RETURN_BUSY_DESC

            refresh_system_language()

            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UpdateCache sender:%s...',sender_name)
            self.parent.start_update_cache()
            return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC
        except Exception as e:
            logging.error(str(e))
            return self.RETURN_UNKNOWN_CODE,str(e)


    # 以下是适配conf2新增接口

    # 有可更新内容弹窗
    # 获取有可更新内容弹窗状态
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b',sender_keyword='sender')
    def GetAutocheckStatus(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetAutocheckStatus sender: %s .', sender_name)

        autocheckStatus = True

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.notification', 'enable')
            autocheckStatus = False if kconf2_key_value.lower() == 'false' else True
        except UpdaterKconf2Err as e:
            autocheckStatus = False if "false" == self.parent.sqlite3_server.select_from_display("auto_check") \
                                    else True
            logging.info("autocheckStatus default: %s.", autocheckStatus)
        except Exception as e:
            logging.error("GetAutocheckStatus: %s.", str(e))

        return autocheckStatus

    # 设置有可更新内容弹窗状态
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='i',sender_keyword='sender')
    def SetAutocheckStatus(self,status,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutocheckStatus sender: %s .', sender_name)

        setResult = int(0)

        try:
            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.notification', 
                                                                'enable', 
                                                                'true' if status else 'false') else int(1)
        except Exception as e:
            logging.error("SetAutocheckStatus: %s.", str(e))
            setResult = int(1)

        return setResult

    # 下载限速与限速值
    # 获取下载限速状态
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b',sender_keyword='sender')
    def GetDownloadLimitStatus(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetDownloadLimitStatus sender: %s .', sender_name)

        downloadspeedLimitStatus = True

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadlimit', 'enable')
            downloadspeedLimitStatus = False if kconf2_key_value.lower() == 'false' else True
        except UpdaterKconf2Err as e:
            downloadspeedLimitStatus = True if "true" == self.parent.sqlite3_server.select_from_display("download_limit") \
                                            else False
            logging.info("downloadspeedLimitStatus default: %s.", downloadspeedLimitStatus)
        except Exception as e:
            logging.error("GetDownloadLimitStatus: %s.", str(e))

        return downloadspeedLimitStatus
    
    # 设置下载限速状态
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='i',sender_keyword='sender')
    def SetDownloadLimitStatus(self,status,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetDownloadLimitStatus sender: %s .', sender_name)

        setResult = int(0)

        try:
            setResult = self.kconf2_set_downloadlimit_enable(status)
        except Exception as e:
            logging.error("SetDownloadLimitStatus: %s.", str(e))
            setResult = int(1)

        return setResult

    # 获取下载限速值
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='i',sender_keyword='sender')
    def GetDownloadLimit(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetDownloadLimit sender: %s .', sender_name)

        downloadlimitValue = int(0)

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadlimit', 'value')
            downloadlimitValue = int(0) if not kconf2_key_value.isdigit() else int(kconf2_key_value)
        except UpdaterKconf2Err as e:
            downloadlimitValue = int(self.parent.sqlite3_server.select_from_display("download_limit_value"))
            logging.info("downloadlimitValue default: %s.", downloadlimitValue)
        except Exception as e:
            logging.error("GetDownloadLimit: %s.", str(e))

        return downloadlimitValue

    # 设置下载限速值
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='i',out_signature='i',sender_keyword='sender')
    def SetDownloadLimit(self,_speed_value,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetDownloadLimit sender: %s .', sender_name)

        setResult = int(0)

        try:
            setResult = self.kconf2_set_downloadlimit_value(_speed_value)
        except Exception as e:
            logging.error("SetDownloadLimit: %s.", str(e))
            setResult = int(1)

        return setResult
    

    # 自动下载更新
    # 获取自动更新状态
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b',sender_keyword='sender')
    def GetAutoUpgradeStatus(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetAutoUpgradeStatus sender: %s .', sender_name)

        autoUpgradeStatus = True

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.autodownload', 'enable')
            autoUpgradeStatus = False if kconf2_key_value.lower() == 'false' else True
        except UpdaterKconf2Err as e:
            autoUpgradeStatus = True if "true" == self.parent.sqlite3_server.select_from_display("autoupdate_allow") \
                                     else False
            logging.info("autoUpgradeStatus default: %s.", autoUpgradeStatus)
        except Exception as e:
            logging.error("GetAutoUpgradeStatus: %s.", str(e))

        return autoUpgradeStatus

    # 设置自动更新状态
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='i',sender_keyword='sender')
    def SetAutoUpgradeStatus(self,status,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutoUpgradeStatus sender: %s .', sender_name)

        setResult = int(0)

        try:
            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.autodownload', 
                                                                'enable', 
                                                                'true' if status else 'false') else int(1)
        except Exception as e:
            logging.error("SetAutoUpgradeStatus: %s.", str(e))
            setResult = int(1)

        return setResult


    # 自动更新配置
    # 获取自动更新下载周期
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='i',sender_keyword='sender')
    def GetAutoupgradePeriod(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetAutoupgradePeriod sender: %s .', sender_name)

        autoupgradePeriod = int(7)

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadsettings', 'period')
            autoupgradePeriod = int(kconf2_key_value) if kconf2_key_value.isdigit() else int(7)
        except UpdaterKconf2Err as e:
            autoupgradePeriod = int(self.parent.sqlite3_server.select_from_display("update_period"))
            logging.info("autoupgradePeriod default: %s.", autoupgradePeriod)
        except Exception as e:
            logging.error("GetAutoupgradePeriod: %s.", str(e))

        return autoupgradePeriod

    # 设置自动更新下载周期
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='i',out_signature='i',sender_keyword='sender')
    def SetAutoupgradePeriod(self,_period,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutoupgradePeriod sender: %s .', sender_name)

        setResult = int(0)

        try:
            setResult = self.kconf2_set_downloadsettings_period(_period)
        except Exception as e:
            logging.error("SetAutoupgradePeriod: %s.", str(e))
            setResult = int(1)

        return setResult

    # 获取自动更新下载策略
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='bs',sender_keyword='sender')
    def GetAutoupgradeDetail(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetAutoupgradeDetail sender: %s .', sender_name)

        designatedStatus = True
        designatedtime   = '18:00-08:00'

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadsettings', 'designated')
            designatedStatus = False if kconf2_key_value.lower() == 'false' else True

            kconf2_key_value1 = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.downloadsettings', 'designatedtime')
            designatedtime = str(kconf2_key_value1)
        except Exception as e:
            logging.error("GetAutoupgradeDetail: %s.", str(e))

        return designatedStatus, designatedtime

    # 设置自动更新下载策略
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='bs',out_signature='i',sender_keyword='sender')
    def SetAutoupgradeDetail(self,status,designatedtime,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutoupgradeDetail sender: %s .', sender_name)

        setResult = int(0)

        try:
            setResult = self.kconf2_set_downloadsettings_designated(status, designatedtime)

            self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadTime", str(designatedtime), True)
            self.PropertyChanged("downloadTime", str(designatedtime))
            self.ButtonStatusChange("autoUpgradeTime", str(designatedtime))
        except Exception as e:
            logging.error("SetAutoupgradeDetail: %s.", str(e))
            setResult = int(1)

        return setResult

    # 自动备份配置
    # 获取是否自动备份
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b',sender_keyword='sender')
    def GetAutoBackupStatus(self,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetAutoBackupStatus sender: %s .', sender_name)

        autoBackupStatus = True

        try:
            kconf2_key_value = self.parent.updater_kconf2.conf2_get_key_value('kylin-update-frontend.autobackup', 'enable')
            autoBackupStatus = False if kconf2_key_value.lower() == 'false' else True
        except Exception as e:
            logging.error("GetAutoBackupStatus: %s.", str(e))

        return autoBackupStatus

    # 设置是否跳过备份
    # true:打开，false:关闭
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='i',sender_keyword='sender')
    def SetAutoBackup(self,status,sender=None):
        sender_name = self.dbus_control.check_sender(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutoBackup sender: %s .', sender_name)

        setResult = int(0)

        try:
            setResult = int(0) if self.parent.updater_kconf2.conf2_set_key_value(
                                                                'kylin-update-frontend.autobackup', 
                                                                'enable', 
                                                                'true' if status else 'false') else int(1)
        except Exception as e:
            logging.error("SetAutoBackup: %s.", str(e))
            setResult = int(1)

        return setResult

    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='b')
    def AutocheckStatusChanged(self, status):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " AutocheckStatusChanged status = %r", status)

    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bi')
    def DownloadStatusChanged(self, status, value):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " DownloadStatusChanged status = %r, value = %s",\
                      status, value)

    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='b')
    def AutoUpgradeStatusChanged(self, status):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " AutoUpgradeStatusChanged status = %r", status)

    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='i')
    def AutoupgradePeriodChanged(self, _period):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " AutoupgradePeriodChanged period = %i", _period)

    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bs')
    def AutoupgradeDetailChanged(self, status, designatedtime):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " AutoupgradeDetailChanged status = %r, designatedtime = %s", status, designatedtime)

    #限速修改信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='ss')
    def ButtonStatusChange(self, signal_types = '', value=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " ButtonStatusChange signal_types = %s, value = %s.",signal_types, value)

    # dbus 信号：用于发送立即更新信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE)
    def UpgradeAllNow(self):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" UpgradeAllNow") 
    
    # dbus 信号：用于发送自动更新配置更改信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE)
    def ChangeUpgradePolicy(self):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" ChangeUpgradePolicy") 


    #限速修改信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='ss')
    def ButtonStatusChange(self, signal_types = '', value=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " ButtonStatusChange signal_types = %s, value = %s.",signal_types, value)


    # signal：属性发生改变
    @dbus.service.signal(dbus_interface=UPDATER_DBUS_INTERFACE,
                         signature="ss")
    def PropertyChanged(self, property, value):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" PropertyChanged: ( %s, %s )" % (property, value))

    # signal：属性发生改变
    @dbus.service.signal(dbus_interface=UPDATER_DBUS_INTERFACE,
                         signature="ss")
    def StrategyChanged(self, property, value):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" StrategyChanged: ( %s, %s )" % (property, value))
    
    def UpdateCacheFinishedNext(self, success, upgrade_group,error_string='',error_desc=''):
        try:
            os.remove(STRATEGIES_UPDATING_CACHE)
        except:
            pass

        self.UpdateCacheFinished(success, upgrade_group, error_string, error_desc)

    #apt cache更新完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='basss')
    def UpdateCacheFinished(self, success, upgrade_group,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " UpdateCacheFinished success = %r , upgrade_group = %a, error_string = %s , error_desc = %s ",\
                    success,upgrade_group,error_string,error_desc)

    #apt cache更新进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='is')
    def UpdateCacheStatusChanged(self,progress,status):
        logging.info(COLORLOG_PREFIX+"Emitting"+COLORLOG_SUFFIX+" UpdatCacheStatusChanged progress = %d , status = %s",progress,status)

    # 信号是否可取消
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='b')
    def Cancelable(self, Cancelable):
        # pass
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" Cancelable: %r",Cancelable)

    # WRITABLE_PROPERTIES = ()

    # # pylint: disable-msg=C0103,C0322
    # @dbus.service.signal(dbus_interface=dbus.PROPERTIES_IFACE,
    #                     signature="sa{sv}as")
    # def PropertiesChanged(self, interface, changed_properties,
    #                     invalidated_properties):
    #     """The signal gets emitted if a property of the object's
    #     interfaces changed.

    #     :param property: The name of the interface.
    #     :param changed_properties: A dictrionary of changed
    #         property/value pairs
    #     :param invalidated_properties: An array of property names which
    #         changed but the value isn't conveyed.

    #     :type interface: s
    #     :type changed_properties: a{sv}
    #     :type invalidated_properties: as
    #     """
    #     logging.info("Emitting PropertiesChanged: %s, %s, %s" %
    #             (interface, changed_properties, invalidated_properties))

    # # pylint: disable-msg=C0103,C0322
    # @dbus.service.method(dbus.INTROSPECTABLE_IFACE,
    #                     in_signature='', out_signature='s',
    #                     path_keyword='object_path',
    #                     connection_keyword='connection')
    # def Introspect(self, object_path, connection):
    #     # Inject the properties into the introspection xml data
    #     data = dbus.service.Object.Introspect(self, object_path, connection)
    #     xml = ElementTree.fromstring(data)
    #     for iface in xml.findall("interface"):
    #         props = self._get_properties(iface.attrib["name"])
    #         for key, value in props.items():
    #             attrib = {"name": key}
    #             if key in self.WRITABLE_PROPERTIES:
    #                 attrib["access"] = "readwrite"
    #             else:
    #                 attrib["access"] = "read"
    #             if isinstance(value, dbus.String):
    #                 attrib["type"] = "s"
    #             elif isinstance(value, dbus.UInt32):
    #                 attrib["type"] = "u"
    #             elif isinstance(value, dbus.Int32):
    #                 attrib["type"] = "i"
    #             elif isinstance(value, dbus.UInt64):
    #                 attrib["type"] = "t"
    #             elif isinstance(value, dbus.Int64):
    #                 attrib["type"] = "x"
    #             elif isinstance(value, dbus.Boolean):
    #                 attrib["type"] = "b"
    #             elif isinstance(value, dbus.Struct):
    #                 attrib["type"] = "(%s)" % value.signature
    #             elif isinstance(value, dbus.Dictionary):
    #                 attrib["type"] = "a{%s}" % value.signature
    #             elif isinstance(value, dbus.Array):
    #                 attrib["type"] = "a%s" % value.signature
    #             else:
    #                 raise Exception("Type %s of property %s isn't "
    #                                 "convertable" % (type(value), key))
    #             iface.append(ElementTree.Element("property", attrib))
    #     new_data = ElementTree.tostring(xml, encoding="UTF-8")
    #     return new_data

    # # pylint: disable-msg=C0103,C0322
    # @dbus.service.method(dbus.PROPERTIES_IFACE,
    #                     in_signature="ssv", out_signature="",
    #                     sender_keyword="sender")
    # def Set(self, iface, name, value, sender):
    #     """Set a property.

    #     Only the user who intiaited the transaction is
    #     allowed to modify it.

    #     :param iface: The interface which provides the property.
    #     :param name: The name of the property which should be modified.
    #     :param value: The new value of the property.

    #     :type iface: s
    #     :type name: s
    #     :type value: v
    #     """
    #     logging.info("Set() was called: %s, %s" % (name, value))
    #     return self._set_property(iface, name, value, sender)

    # # pylint: disable-msg=C0103,C0322
    # @dbus.service.method(dbus.PROPERTIES_IFACE,
    #                     in_signature="s", out_signature="a{sv}")
    # def GetAll(self, iface):
    #     """Get all available properties of the given interface."""
    #     logging.info("GetAll() was called: %s" % iface)
    #     return self._get_properties(iface)

    # # pylint: disable-msg=C0103,C0322
    # @dbus.service.method(dbus.PROPERTIES_IFACE,
    #                     in_signature="ss", out_signature="v")
    # def Get(self, iface, property):
    #     """Return the value of the given property provided by the given
    #     interface.
    #     """
    #     logging.info("Get() was called: %s, %s" % (iface, property))
    #     return self._get_properties(iface)[property]
