#!/usr/bin/env python
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

"""Integration of package managers into SystemUpdater"""

import os
import json
import logging
import subprocess
import traceback
import dbus
import shutil
import hashlib
import urllib.parse
import time
from gettext import gettext as _
from SystemUpdater.Core.errors import *
from SystemUpdater.Core.enums import *
from SystemUpdater.Core.PluginManager import *
from SystemUpdater.Core.UpdateEssentialItem import remove_essential
from SystemUpdater.Core.PlymouthManager import PlymouthManager
from SystemUpdater.Core.shutdown_lock import InhibitShutdownLock
from SystemUpdater.Core.space_rules import check_free_space
from aptdaemon.enums import ERROR_PACKAGE_DOWNLOAD_FAILED,ERROR_PACKAGE_MANAGER_FAILED,ERROR_NO_PACKAGE
from SystemUpdater.Core.BackupProperties import backup_serverid
from SystemUpdater.Core.GenericConfigParser import Gconfig
from SystemUpdater.Core.utils import print_now_sources
from SystemUpdater.Core.power_manager import check_device_power

class AptdBaseInformation:
    def __init__(self,
                progress = 1,
                status='',
                details='',
                cancelable = False,
                error_details = '',
                current_items = 1,
                total_items = 1,
                on_installing = False,
                retry_limit = 2
                ):

        self._progress = progress
        self.status = status
        self.details = details
        self.cancelable = cancelable
        self.error_details = error_details
        self.current_items = current_items
        self.total_items = total_items
        self.on_installing = on_installing
        self.retry_limit = retry_limit

        self.start_progress = 0
        self.end_progress = 100

    def _get_progress(self):
        return self._progress

    def _set_progress(self, progress):
        if self.current_items == 0 and \
                self.total_items == 0 and self.on_installing == False:
            return

        if self.start_progress == 0 and self.end_progress == 100:
            pass
        else:
            progress = int(self.start_progress + (progress / 100) *
                        (self.end_progress - self.start_progress))
        
        # 防止出现 计算出的进去为0的情况
        if progress > self._progress:
            self._progress = progress

    progress = property(_get_progress, _set_progress,
                    doc="The required disk space of the transaction.")

class InstallBackend(GObject.GObject):

    __gsignals__ = {"finished": (GObject.SIGNAL_RUN_FIRST,
                                GObject.TYPE_NONE,
                                (   GObject.TYPE_BOOLEAN,
                                    GObject.TYPE_STRING
                                )),
                    }

    ACTION_DEFUALT_STATUS = -1
    ACTION_UPDATE = 0
    ACTION_INSTALL = 1
    ACTION_INSTALL_DEB = 2
    ACTION_CHECK_RESOLVER = 3
    ACTION_DOWNLOADONLY = 4
    ACTION_FIX_BROKEN = 5
    ACTION_REMOVE_PACKAGES = 6
    ACTION_FIX_INCOMPLETE = 7
    ACTION_CLEAN = 8
    ACTION_INSTALL_SHUTDOWN = 9
    ACTION_BACKGROUND_UPGRADE = 10
    ACTION_INSTALL_DEB_RESOLVER = 11

    MODE_DEFAULT_STATUS = -1
    
    MODE_INSTALL_PARTIAL = 0
    MODE_INSTALL_ALL = 1
    MODE_INSTALL_SYSTEM = 2

    MODE_UPDATE_CACHE = 0
    MODE_UPDATE_ALL = 1
    MODE_SLEF_UPGRADE = 2

    MODE_OTHER = 1
    MODE_BACKGROUND_UPGRADE = 2
    MODE_BACKGROUND_UPDATE_CACHE = 3
    MODE_BACKGROUND_UPDATE_INSTALL = 4
    MODE_OTHER_UPDATE_CACHE = 5

    def __init__(self, window_main, action,mode = MODE_DEFAULT_STATUS,data = None):
        GObject.GObject.__init__(self)
        self.window_main = window_main
        self.bus = window_main.bus
        self.cache = window_main.cache
        self.action = action
        self.action_mode = mode

        self.window_main.now_working = action

        self.aptd_base = AptdBaseInformation()
        self.collector = window_main.collector

        self.backend_data = None
        self.upgrading_content = []
        if data:
            self.backend_data = data
            # 安装时 将上次和本次全部的更新内容进行安装升级
            if self.action == self.ACTION_INSTALL_SHUTDOWN:
                self.upgrading_content = self.backend_data.get_actual_updated_content()
            else:
                # 下载时 仅上报下载的内容
                self.upgrading_content = self.backend_data.content_for_frontend

        self.dbus_send = window_main.dbusController
        self.database = window_main.sqlite3_server
        self.inhibit_shutdown = InhibitShutdownLock()

        self.plymouth = PlymouthManager()
        
        self.scripts = self.window_main.scripts_manager
        
    def start(self):
        try:
            if self.action == self.ACTION_UPDATE and self.action_mode == self.MODE_SLEF_UPGRADE:
                pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade = self.get_mark_from_cache(self.cache)

                logging.info("SELF-UPGRADE install:%d , upgrade:%d remove:%d downgrade:%d",len(pkgs_install),\
                                                    len(pkgs_upgrade),len(pkgs_remove),len(pkgs_downgrade))
                
                remove_essential(pkgs_remove)
                self.commit_next(self.action,pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_downgrade)
            #下载前的准备
            elif self.action == self.ACTION_DOWNLOADONLY:
                #0、准备数据
                pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade = self.backend_data.get_resolver_upgrade()
                self.aptd_base.end_progress = self.backend_data.get_download_progress()

                if len(pkgs_install) == 0 and len(pkgs_upgrade) == 0 and len(pkgs_downgrade) == 0:
                    raise UpdateBaseError(ERROR_NOT_UPGRADE_PACKAGES)
                    
                logging.info("INSTALL install:%d , upgrade:%d remove:%d,downgrade:%d",\
                                                len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove),len(pkgs_downgrade))
                
                self.commit_download(pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_downgrade)
            elif self.action == self.ACTION_INSTALL or self.action == self.ACTION_INSTALL_SHUTDOWN:
                if self.action == self.ACTION_INSTALL_SHUTDOWN:
                    self.plymouth.show()
                    mesa = _("The system is installing updates. Please do not turn off your computer. ")
                    self.plymouth.message_to(mesa)
                    self.plymouth.progress_to(1)

                    # 标记需要还原，当安装过程断电，或者安装失败的情况下自动还原
                    if self.backend_data.restorable:
                        self.inhibit_shutdown.mark_restore()
                    self.echo_detial_log_to_tty()
                
                # HWE的BUG328201，当升级为2303-update2->2303-update2质量更新3.0->2503，因为某种原因，导致当前缓存的包是错误的，安装失败
                # 处理方法：实测包数量较少，个位数，在安装前进行缓存的包校验，如果校验失败直接删除包，在安装阶段重新下载
                start_time = time.time()
                self._check_debs_md5(self.cache)
                end_time = time.time()
                execution_time = end_time - start_time  # 计算执行时间
                logging.info(f"_check_debs_md5 function took {execution_time} seconds")


                for scr in self.backend_data.get_pre_script():
                    self.scripts.run_plugin(scr)

                self._get_install_lock(_("Kylin System Updater"),inhibit_sleep = True)

                logging.info("The task Queue(%r) to be upgraded...",self.backend_data.steps_queue)
                logging.info("Final State(%s) of the device after installation is completed...",self.backend_data.device_status)
                pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade = self.backend_data.get_resolver_steps(self.backend_data.resolver_name)

                self._upgrade_content_for_safe(pkgs_install,pkgs_upgrade,pkgs_remove)
                self._start_install_queue()
            elif self.action == self.ACTION_FIX_BROKEN:
                pkgs_install,pkgs_upgrade,pkgs_remove = self._get_pkgs_from_cache(self.cache)

                remove_essential(pkgs_remove)
                self.fix_broken()
            elif self.action == self.ACTION_FIX_INCOMPLETE:
                self.fix_incomplete()
            elif self.action == self.ACTION_CLEAN:
                self.clean()
            elif self.action == self.ACTION_UPDATE:
                check_free_space()
                self.update()
        except UpdateBaseError as excep:
            self._action_done(self.action,False,excep.header,excep.desc,excep.code)
        except UpdateProgressExit as excep:
            pass
        except Exception as e:
            logging.error(e) 
            tbk = traceback.format_exc()
            logging.error(str(tbk))
        
    def start_alone(self,push_content = [],_is_install = False, caller=''):
        # 安装本地deb包的接口
        self.upgrading_content = push_content
        
        if self.action == self.ACTION_INSTALL_DEB: 
            try:
                self._get_install_lock(caller=caller)
                self.install_deb(install_path = push_content, install_force = _is_install)
            except Exception as e:
                logging.error(str(e))
        elif self.action == self.ACTION_REMOVE_PACKAGES:
            self._get_install_lock(_("Kylin System Updater"))
            self.purge_packages(push_content)
        elif self.action == self.ACTION_INSTALL_DEB_RESOLVER:
            try:
                self.window_main.refresh_cache(False,True)
                pkgs_install,pkgs_upgrade,pkgs_remove,x = self.window_main.cache.make_problem_resolver(pkgs_install = push_content)
                self._get_install_lock(caller=caller)
                self.commit(self.ACTION_INSTALL_DEB_RESOLVER,pkgs_install,pkgs_upgrade,pkgs_remove)
            except Exception as e:
                logging.error(str(e))
        elif self.action == self.ACTION_BACKGROUND_UPGRADE and self.action_mode == self.MODE_OTHER_UPDATE_CACHE:
            check_free_space()
            self.update_local()
        elif self.action == self.ACTION_BACKGROUND_UPGRADE and self.action_mode == self.MODE_BACKGROUND_UPDATE_CACHE:
            check_free_space()
            self.update()
        elif self.action == self.ACTION_BACKGROUND_UPGRADE:
            try:
                pkgs_installs                = []
                pkgs_upgrade                 = []
                pkgs_remove                  = []
                pkgs_install_with_version    = []
                pkgs_install_without_version = []

                pkgs_install = [ str(pkg) for pkg in push_content]
                if self.action_mode != self.MODE_BACKGROUND_UPDATE_INSTALL:
                    self.window_main.pkgs_install_bak = pkgs_install
                else:
                    pkgs_install = self.window_main.pkgs_install_bak

                logging.info("Install deb package, open cache %d", self.action_mode)
                if self.action_mode == self.MODE_BACKGROUND_UPGRADE:
                    self.window_main.refresh_cache(True,True)
                elif self.action_mode == self.MODE_OTHER or self.action_mode == self.MODE_BACKGROUND_UPDATE_INSTALL or self.action_mode == self.MODE_OTHER_UPDATE_CACHE:
                    self.window_main.refresh_cache(False,True)
                for pkg_name in pkgs_install:
                    if pkg_name not in self.window_main.cache:
                        if "=" not in pkg_name or ("=" in pkg_name and pkg_name.split("=")[0] not in self.window_main.cache):
                            if self.action_mode == self.MODE_BACKGROUND_UPGRADE:
                                self.window_main.start_back_update_cache(InstallBackend.ACTION_BACKGROUND_UPGRADE, InstallBackend.MODE_BACKGROUND_UPDATE_CACHE)
                            if self.action_mode == self.MODE_OTHER:
                                logging.warning("'%s' is not in cache, need update cache and retry", pkg_name)
                                self.window_main.start_back_update_cache(InstallBackend.ACTION_BACKGROUND_UPGRADE, InstallBackend.MODE_OTHER_UPDATE_CACHE)
                            else:
                                logging.error("'%s' is not in cache", pkg_name)
                                self._action_done(self.action,False,"","Package is not in cache",ERROR_NO_PACKAGE_CACHE)
                            return
                    
                    if "=" in pkg_name:
                        pkgs_install_with_version.append(pkg_name)
                    else:
                        pkgs_install_without_version.append(pkg_name)

                if (len(pkgs_install_without_version)!=0):
                    pkgs_installs,pkgs_upgrade,pkgs_remove,x = self.window_main.cache.make_resolver_self_upgrade(pkgs_install_without_version)
                    logging.info("background install list: "+",".join(pkgs_installs)+" , *** upgrade list: "+",".join(pkgs_upgrade))

                    try:
                        # 检查磁盘空间，出现任何错误不阻塞流程
                        required_download = getattr(self.window_main.cache, "required_download",0)
                        additional_required_space = getattr(self.window_main.cache, "additional_required_space",0)
                        check_free_space(required_download,additional_required_space)

                        # 检查删除重要软件包
                        remove_essential(pkgs_remove)

                        if caller == "ksc-vulnerability-repair-daemon":
                            check_device_power(self.bus,len(pkgs_installs)+len(pkgs_upgrade)+len(pkgs_remove))
                    except UpdateBaseError as excep:
                        self._action_done(self.action,False,"","",excep.code)
                        return
                    except Exception as e:
                        logging.info(str(e))
                self._get_install_lock(caller=caller)
                if self.action_mode == self.MODE_OTHER or self.action_mode == self.MODE_BACKGROUND_UPDATE_INSTALL or self.action_mode == self.MODE_OTHER_UPDATE_CACHE:
                    self.commit(self.ACTION_BACKGROUND_UPGRADE,pkgs_installs+pkgs_install_with_version,pkgs_upgrade,pkgs_remove)
                else:
                    self.commit_next(self.ACTION_BACKGROUND_UPGRADE,pkgs_installs+pkgs_install_with_version,pkgs_upgrade,pkgs_remove)
            except UpdateBaseError as e:
                start_index = None
                simulate_result = ""
                if e.code == ERROR_RESOLVER_FAILED:
                    args = ["apt-get","install","--simulate"]
                    args = args + pkgs_install
                    p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)

                    lines = p.stdout.splitlines()
                    for i,line in enumerate(lines):
                        if "下列软件包有未满足的依赖关系" in line:
                            start_index = i
                            break
                    if start_index is not None:
                        simulate_result = lines[start_index:]
                    logging.error("InstallPackages simulate install: %s.", " ".join(simulate_result) if isinstance(simulate_result, list) else "")
                self._action_done(self.action,False,str(e)," ".join(simulate_result) if isinstance(simulate_result, list) else "",ERROR_RESOLVER_FAILED)
            except Exception as e:
                self._action_done(self.action,False,str(e),"","")

    def update(self):
        """Run a update to refresh the package list"""
        raise NotImplementedError

    def clean(self):
        """Remove all downloaded files.t"""
        raise NotImplementedError

    def commit(self,model,pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_downgrade = []):
        """Commit the cache changes """
        raise NotImplementedError

    def install_deb(self,install_path,install_force):
        """install_deb the cache changes """
        raise NotImplementedError

    def fix_broken(self):
        """fix_broken_depends"""
        raise NotImplementedError

    def fix_incomplete(self):
        """fix_incomplete_install"""
        raise NotImplementedError

    def purge_packages(self,pkgs_purge):
        """purge_packages"""
        raise NotImplementedError

    def _get_pkgs_from_cache(self,cache):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []
        for pkg in cache:
            try:
                if pkg.marked_install :
                    pkgs_install.append(pkg.name)
                elif pkg.marked_upgrade:
                    pkgs_upgrade.append(pkg.name)
                elif pkg.marked_delete:
                    pkgs_remove.append(pkg.name)
            except KeyError:
                # pkg missing from fresh_cache can't be modified
                pass
        return pkgs_install,pkgs_upgrade,pkgs_remove

    def get_mark_from_cache(self,cache):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []
        pkgs_downgrade = []
        
        for pkg in cache:
            try:
                if pkg.marked_install:
                    pkgname = pkg.name
                    if pkg.is_auto_installed:
                        pkgname += "#auto"
                        pkgs_install.append(pkgname)
                    else:
                        pkgs_install.append(pkgname)
                elif pkg.marked_upgrade:
                        pkgs_upgrade.append(pkg.name)
                elif pkg.marked_delete:
                    pkgs_remove.append(pkg.name)
                elif pkg.marked_downgrade:
                    pkgs_downgrade.append(pkg.name+"="+getattr(pkg.candidate, "version", ''))
            except KeyError:
                pass
        return pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade

    #检查是否需要重新下载
    def _need_retry_download(self):
        def _check_connectivity_pro():
            try:
                logging.info("Start checking the network...")
                obj = self.bus.get_object("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager")
                interface = dbus.Interface(obj, "org.freedesktop.NetworkManager")
                retval = interface.CheckConnectivity(timeout=15)
            except Exception as e:
                logging.error("Network status is Exception:" + str(e))
                return False
            if retval == 4:
                logging.info("Network status is Success...")
                return True
            else:
                logging.info("Network status(%d) is Exception...",retval)
                return False
        
        if _check_connectivity_pro() == True and self.aptd_base.retry_limit != 0:
            self.aptd_base.retry_limit = self.aptd_base.retry_limit - 1
            logging.warning("Retry the upgrade and installaton and retry number: %d",self.aptd_base.retry_limit)
            pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade = self.backend_data.get_resolver_upgrade()
            self.commit_download(pkgs_install, pkgs_upgrade, pkgs_remove)
            return True
        else:
            return False

    def _need_retry_install(self):
        if self.aptd_base.retry_limit != 0:
            self.aptd_base.retry_limit = self.aptd_base.retry_limit - 1
            logging.warning("Retry the upgrade and installaton and retry number: %d",self.aptd_base.retry_limit)
            pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade = self.backend_data.get_resolver_steps(self.backend_data.last_task)

            if self.backend_data.check_steps_reboot(self.backend_data.last_task) or self.backend_data.steps_queue == []:
                final_stage = True

            logging.info("INSTALL(%s) install:%d , upgrade:%d remove:%d",self.backend_data.last_task,len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
            self.commit_install(pkgs_install, pkgs_upgrade, pkgs_remove,final_stage)
            return True

    def _start_install_queue(self):
        final_stage = False

        now_steps = self.backend_data.steps_queue.pop()
        self.aptd_base.start_progress,self.aptd_base.end_progress = self.backend_data.get_install_progress(now_steps)
        pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade = self.backend_data.get_resolver_steps(now_steps)

        self.dbus_send.NotifyStepsInstalled(True,self.backend_data.get_mid_notify(now_steps))

        if self.backend_data.check_steps_reboot(now_steps) or self.backend_data.steps_queue == []:
            final_stage = True

        self.backend_data.last_task = now_steps

        logging.info("INSTALL(%s) install:%d , upgrade:%d remove:%d,downgrade:%d",now_steps,len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove),len(pkgs_downgrade))
        self.commit_install(pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_downgrade,final_stage)

    def _action_done(self, action,success, error_string='',error_desc='',error_code=''):
        if error_code == ERROR_NO_PACKAGE:
            import apt_pkg
            # 在出现异常的情况下 保存源与索引文件 方便定位问题
            if os.path.isdir("/var/lib/kylin-system-updater/lists"):
                shutil.make_archive("/var/cache/kylin-system-updater/lists.bak", 'tar', "/var/lib/kylin-system-updater/lists")
            print_now_sources()

        self.window_main.now_working = self.ACTION_DEFUALT_STATUS
        if self.action == self.ACTION_UPDATE and self.action_mode == self.MODE_SLEF_UPGRADE:
            self.emit("finished",success,error_code)
        elif action == self.ACTION_DOWNLOADONLY:
            #1、检查是否需要重试下载
            if success == False and error_code == ERROR_PACKAGE_DOWNLOAD_FAILED:
                if self._need_retry_download():
                    return
            self.emit("finished",success,error_code)
            
        elif action == self.ACTION_INSTALL or action == self.ACTION_INSTALL_SHUTDOWN:            
            # 在离线更新特定机型下 会出现离线源下载失败的问题
            if success == False and self.aptd_base.error_details != "" and \
                                        "(5: Input/output error)" in self.aptd_base.error_details:
                self._need_retry_install()
            if action == self.ACTION_INSTALL_SHUTDOWN:
                # bug:328381 【在线更新】关机安装界面自动生成安装日志时，安装完成时日志不会和进度条同时消失
                self.plymouth.message_to("updater:")
            #1、解除安装状态
            self._release_install_lock()

            #2、保存更新状态到系统，提供外部使用
            if success:
                self._set_update_flag(self.upgrading_content)
            
            # 安装成功或者其他安装错误不还原，仅安装失败的情况还原
            if error_code not in (ERROR_PACKAGE_MANAGER_FAILED,ERROR_INSTALL_TIMEOUT_FAILED,ERROR_INSTALL_POWER_OUTAGE):
                self.inhibit_shutdown.remove_restore_flag()

            if error_code == ERROR_PACKAGE_MANAGER_FAILED:
                self._output_error_dsc()

            #执行安装后脚本
            for scr in self.backend_data.get_post_script():
                self.scripts.run_plugin(scr)
                        
            self.emit("finished",success,error_code)
        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_ALL:
            backup_serverid()
            self.emit("finished",success,error_code)
        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_CACHE:
            backup_serverid()
            self.dbus_send.UpdateCacheFinished(success,[''],error_string,error_desc)
        elif action == self.ACTION_FIX_BROKEN or action == self.ACTION_FIX_INCOMPLETE:
            self.emit("finished",success,error_code)

        elif action == self.ACTION_REMOVE_PACKAGES:
            self._release_install_lock()
            purge_list = self.database.current_purge_pkgs
            if success == False and 'in the prevent uninstallation of blacklists' in self.aptd_base.error_details \
                and error_code == 'error-package-manager-failed':
                error_string = _("Forbid uninstallation of software packages.")
                error_desc   = _("Package is in the prevent uninstallation of blacklists!")
            if success == False and 'not in the prevent uninstallation of whitelist!' in self.aptd_base.error_details \
                and error_code == 'error-package-manager-failed':
                error_string = _("Forbid uninstallation of software packages.")
                error_desc   = _("Package is not in the prevent uninstallation of whitelist!")

            if str(error_code) == ERROR_NO_PACKAGE:
                success = True

            self.dbus_send.PurgePackagesFinishedWithPkgname(success,error_string,error_desc,",".join(purge_list))
            self.dbus_send.PurgePackagesFinished(success,error_string,error_desc)
            self.dbus_send.PurgePackagesFinishedWithErrCode(success,error_string,error_desc,",".join(purge_list),get_error_num_from_enum(error_code))
            
        elif action == self.ACTION_INSTALL_DEB or action == self.ACTION_INSTALL_DEB_RESOLVER:
            self._release_install_lock()
            if success == False and '\r\n: \r\n\r\n' in self.aptd_base.error_details:
                error_string = _("Package validation failed and installation was rejected.")
                error_desc = ''
            if error_code != None and str(error_code) != 'error-dep-resolution-failed':
                if error_desc != '':
                    error_string = ",".join((error_string,error_desc))
                self.dbus_send.InstalldebFinished(success,error_string,'')
                self.dbus_send.InstalldebFinishedWithErrCode(success,error_string,'',get_error_num_from_enum(error_code))
                UpdateMsg = {}
                if success:
                    status = 'success'
                    UpdateMsg.update({"errorCode":" "})
                else:
                    status = 'failed'
                    UpdateMsg.update({"errorCode":str(error_string+" "+error_desc)})
                #apt发送数据
                if self.window_main.configs_uncover.getWithDefault("SystemStatus", "upload_installer_log", False) == True:
                    
                    UpdateMsg.update({"appname":str(self.window_main.deb_obj.get("debname","None").split("_")[0])})
                    UpdateMsg.update({"source":str(self.window_main.deb_obj.get("source","kylin-system-updater"))})
                    UpdateMsg.update({"status":str(status)})
                    UpdateMsg.update({"new_version":str(self.window_main.deb_obj.get("debname","None").split("_")[1])})
                    UpdateMsg.update({"old_version":str(self.window_main.deb_obj.get("old_version","None"))})
                    self.collector.Upgrade_Process_Msg(action, UpdateMsg.copy())
                    self.window_main.deb_obj = {}
        elif (action == self.ACTION_BACKGROUND_UPGRADE and self.action_mode == self.MODE_BACKGROUND_UPDATE_CACHE) or \
                (action == self.ACTION_BACKGROUND_UPGRADE and self.action_mode == self.MODE_OTHER_UPDATE_CACHE):
            install_backend = get_backend(self.window_main, self.ACTION_BACKGROUND_UPGRADE, self.MODE_BACKGROUND_UPDATE_INSTALL)
            install_backend.start_alone(push_content = self.window_main.pkgs_install_bak)
        elif action == self.ACTION_BACKGROUND_UPGRADE:
            self._release_install_lock()
            UpdateMsg = {}
            if error_code != "":
                error_string = get_error_string_from_enum(error_code)
                error_desc = get_error_description_from_enum(error_code)

            if self.action_mode == self.MODE_BACKGROUND_UPGRADE or self.action_mode == self.MODE_BACKGROUND_UPDATE_INSTALL:
                if success:
                    status = 'success'
                else:
                    status = 'failed'
                self.collector.Generate_Msg(self.collector.background_list, self.action_mode)
                for bl in self.collector.background_list:
                    pkg = self.collector.cache[bl]
                    UpdateMsg.update({"appname":str(bl)})
                    UpdateMsg.update({"source":"Kylin Background upgrade"})
                    UpdateMsg.update({"status":str(status)})
                    UpdateMsg.update({"errorCode":str(error_string+" "+error_desc)})
                    if self.collector.background_version[bl]:
                        UpdateMsg.update({"old_version":self.collector.background_version[bl]})
                    UpdateMsg.update({"new_version":self.collector.cache[bl].candidate.source_version})
                    self.collector.Upgrade_Process_Msg(self.action, UpdateMsg.copy())
                    UpdateMsg.clear()

            self.dbus_send.UpdateInstallFinished(success,self.upgrading_content,get_error_num_from_enum(error_code),error_string+error_desc)
            self.collector.background_list.clear()

    def _set_update_flag(self,update_content):
        file_path = '/tmp/update-status'
        update_data = ""
        if update_content:
            update_data = str(update_content).replace('[', '').replace(']', '')
            update_data = update_data.replace("'", '').replace(',', '')

        with open(file_path, 'w') as file:
            file.write(update_data)

    def _get_install_lock(self, caller='kylin-system-updater',inhibit_sleep = False):
        if caller == "":
            caller = 'kylin-system-updater'
        self.inhibit_shutdown.lock(caller=caller,inhibit_sleep = inhibit_sleep)
        self.aptd_base.on_installing = True

    def _release_install_lock(self):
        self.inhibit_shutdown.unlock()

    def _output_error_dsc(self):
        try:
            args = ["/usr/share/kylin-system-updater/scripts/analyse-update-error.sh","--output"]
            p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            if p.returncode == 0:
                return True
            else:
                return False
        except Exception as exc:
            logging.error(exc)

    def enable_next_steps(self):
        args = ["systemctl","enable","kylin-upgrade-poweroff.service"]
        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        if p.returncode == 0:
            logging.info("start enable kylin-upgrade-poweroff.service to success...")
            return True
        else:
            logging.info("start enable kylin-upgrade-poweroff.service to failed...")
            return False

    def disable_next_steps(self):
        args = ["systemctl","is-enabled","kylin-upgrade-poweroff.service"]
        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        if p.returncode != 0:
            pass
        else:
            logging.info("kylin-upgrade-poweroff.service has been enabled and need to disable...")
            args = ["systemctl","disable","kylin-upgrade-poweroff.service"]
            p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            if p.returncode == 0:
                logging.info("start disabled kylin-upgrade-poweroff.service to success...")
                return True
            else:
                logging.info("start disabled kylin-upgrade-poweroff.service to failed...")
                return False

    def _upgrade_content_for_safe(self,pkgs_install,pkgs_upgrade,pkgs_remove):
        INSTALL_OUTPUT_JSON = "/var/lib/kylin-system-updater/json/requires_upgrade_meta.json"
        try:
            output_upgrade = {}

            output_upgrade.update({"pkgs_install":pkgs_install})
            output_upgrade.update({"pkgs_upgrade":pkgs_upgrade})
            output_upgrade.update({"pkgs_remove":pkgs_remove})

            with open(INSTALL_OUTPUT_JSON, 'w', encoding='utf-8') as f:
                json.dump(output_upgrade, f, ensure_ascii=False, indent=4)

            logging.info("Generate Jsonfile(%s) to complete... ",INSTALL_OUTPUT_JSON)
            logging.info("Update Packages list to config file...")
        except Exception as exc:
            logging.error(exc)

    def echo_detial_log_to_tty(self):
        try:
            ECHO_DETAILS_LOG_DEST = "/dev/tty63"
            if SHOW_DETAILES_CHECK_FLAG in Gconfig().get("skip_check_list",[]):
                return

            command = [
                "tail", "-f", "/var/log/kylin-system-updater/kylin-system-updater.log.1"
            ]
            with open(ECHO_DETAILS_LOG_DEST, "w") as output:
                # 将命令放入后台运行
                process = subprocess.Popen(command, stdout=output, stderr=subprocess.STDOUT)
        except Exception as exc:
            logging.error(exc)

    def _check_debs_md5(self, cache):
        # 运行MD5计算线程
        class MD5CalculatorThread():
            def __init__(self, cache, directory="/var/cache/apt/archives", log_dir="/var/cache/kylin-system-updater/install-debs-md5.json"):
                threading.Thread.__init__(self)
                self.directory = directory
                self.log_dir = log_dir
                self.cache = cache
                self.need_redownload = False
            
            def calculate_md5(self, file_path, buffer_size=1024 * 1024):
                '''
                计算文件的MD5值。
                buffer_size: 每次读取的块大小, 默认为1024*1024 KB
                '''
                md5_hash = hashlib.md5()
                try:
                    with open(file_path, "rb") as f:
                        for byte_block in iter(lambda: f.read(buffer_size), b""):
                            md5_hash.update(byte_block)
                    return md5_hash.hexdigest()
                except Exception as e:
                    logging.error(f"Error reading file {file_path}: {e}")
                    return None

            def run(self):
                '''
                线程运行时调用的函数，用于计算并记录 .deb 文件的 MD5
                '''
                deb_files_md5 = {}
                for root, dirs, files in os.walk(self.directory):
                    for file in files:
                        if file.endswith(".deb"):
                            file_path = os.path.join(root, file)
                            debname = (file.split("_"))[0]
                            debversion = (file.split("_"))[1]
                            #文件名称不允许存在符号，因此本地文件会将符号转换为url编码，需要转换
                            debversion  = urllib.parse.unquote(debversion)
                            # 检查包是否存在于APT缓存中
                            if debname not in cache:
                                logging.warning(f"package {debname} not in apt cache, maybe need apt update")
                                continue
                            package = cache[debname]
                            # 检测当前版本是否和候选版本相同
                            if debversion != package.candidate.version:
                                logging.error(f"package {debname} download file version {debversion} is not candidate{package.candidate.version}")
                                os.remove(file_path)
                                self.need_redownload = True
                                continue
                            # 检测当前下载文件是否和apt索引缓存中校验值相同
                            refmd5 = package.candidate.md5
                            calmd5 = self.calculate_md5(file_path)
                            deb_files_md5[debname] = calmd5
                            if refmd5 != calmd5:
                                logging.error(f"the package {debname} check md5 failed, calmd5 is {calmd5}, refmd5 is {refmd5}, file is {file_path}")
                                os.remove(file_path)
                                self.need_redownload = True
                                continue
                # 存储结果到日志文件
                if self.log_dir is not None:
                    try:
                        with open(self.log_dir, 'w') as file:
                            json.dump(deb_files_md5, file, indent=4)
                            logging.info(f"Md5 logs '{self.log_dir}' has been written.")
                    except Exception as e:
                        logging.error(f"An error occurred while writing log: {e}")

        # 取消掉线程，因为计算md5比较快，实测按照秒为单位，1g大概3s多
        try:
            md5_thread = MD5CalculatorThread(cache)
            md5_thread.run()
        except Exception as e:
            logging.error(f"An error occurred while calculating md5: {e}")
        if md5_thread.need_redownload is True:
            logging.info("debs check failed")
        else:
            logging.warning("debs check ok")
    
# try aptdaemon
if os.path.exists("/usr/sbin/aptd") \
   and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ:
    # check if the gtkwidgets are installed as well
    try:
        from .InstallBackendAptdaemon import InstallBackendAptdaemon
    except ImportError:
        logging.exception("importing aptdaemon")

def get_backend(*args, **kwargs):
    """Select and return a package manager backend."""
    # try aptdaemon
    if (os.path.exists("/usr/sbin/aptd")
            and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ):
        # check if the gtkwidgets are installed as well
        try:
            return InstallBackendAptdaemon(*args, **kwargs)
        except NameError:
            logging.exception("using aptdaemon failed")

    # nothing found, raise
    raise Exception("No working backend found, please try installing "
                    "aptdaemon or synaptic")